from PyQt6.QtCore import QAbstractTableModel, Qt
from PyQt6.QtWidgets import QCommonStyle, QStyle

import pixie16


class BinaryDataModel(QAbstractTableModel):
    """Map list mode data to single values that are shown in a row/colum in the UI.

    Data is stored in a list of pypixie.read.Events.

    This class is needed to map entries from the already parsed binary
    data to cells in a QT table.

    This can also be used to highlight or otherwise format the table
    cells as a function of the content.

    """

    def __init__(self, data):
        super().__init__()
        self._data = data
        self.header = pixie16.read.Event.__struct_fields__

        # use some iconds for certain boolean data we also use these
        # to indicate if trace data exist, sine we don't want to
        # show the whole trace in the table
        self.icon_true = QCommonStyle().standardIcon(
            QStyle.StandardPixmap.SP_DialogApplyButton
        )
        self.icon_false = QCommonStyle().standardIcon(
            QStyle.StandardPixmap.SP_DialogCancelButton
        )

    def icon_return(self, value):
        """Return a nice icon for True or False"""
        if value:
            return self.icon_true
        return self.icon_false

    def data(self, index, role):
        """Return the data.

        For some True/False data, we display an icon instead. To do
        this, we return no data in DisplayRole, but an icon for
        DecorationRole.

        """
        column = index.column()
        row = index.row()

        value = getattr(self._data[row], self.header[column])

        if role == Qt.ItemDataRole.DisplayRole:
            if self.header[column] in [
                "trace",
                "CFD_error",
                "pileup",
                "trace_flag",
            ]:
                return None
            return str(value)
        if role == Qt.ItemDataRole.DecorationRole:
            if self.header[index.column()] == "trace":
                return self.icon_return(len(value))
            if self.header[index.column()] in ["CFD_error", "pileup", "trace_flag"]:
                return self.icon_return(not value)

    def rowCount(self, index):
        return len(self._data)

    def columnCount(self, index):
        return len(self.header)

    def headerData(self, section, orientation, role):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self.header[section].title()



import itertools

import numpy as np
from rich import print


from .settings_class import Settings


def compare_dict(A, B, key_string):
    """Deep ompare two dictionarys.

    Return a list of key and values that are different, e.g.
    [...,
     ["a/b/c/", valueA, valueB],
     ...
    ]

    """
    all_keys = set(A.keys()) | set(B.keys())

    output = []

    for key in all_keys:
        new_key_string = f"{key_string}/{key}"
        valueA = A.get(key)
        valueB = B.get(key)
        if isinstance(valueA, dict) and isinstance(valueB, dict):
            output += compare_dict(A[key], B[key], new_key_string)
        elif isinstance(valueA, list) and isinstance(valueB, list):
            output += compare_list(A[key], B[key], new_key_string)
        elif isinstance(valueA, (str, int, float)) and isinstance(
            valueB, (str, int, float)
        ):
            if valueA == valueB:
                pass
            else:
                output.append([new_key_string, valueA, valueB])
        else:
            print(
                f"[orange3]Warning[/] Unkown data types {type(valueA)} {type(valueB)}"
            )
    return output


def compare_list(A, B, key_string):
    """Deep ompare two lists.

    Return a list of key and values that are different, e.g.
    [...,
     ["a/b/c/", valueA, valueB],
     ...
    ]

    """

    output = []

    for i, (valueA, valueB) in enumerate(itertools.zip_longest(A, B)):
        new_key_string = f"{key_string}[{i}]"
        if isinstance(valueA, dict) and isinstance(valueB, dict):
            output += compare_dict(valueA, valueB, new_key_string)
        elif isinstance(valueA, list) and isinstance(valueB, list):
            output += compare_list(valueA, valueB, new_key_string)
        elif isinstance(valueA, (str, int, float)) and isinstance(
            valueB, (str, int, float)
        ):
            if valueA == valueB:
                pass
            else:
                output.append([new_key_string, valueA, valueB])
        else:
            print(
                f"[orange3]Warning[/] Unkown data types {type(valueA)} {type(valueB)}"
            )
    return output


def compare_module_setting(A, B, quiet=False, channels=None):
    """Compare the settings of two modules (for the new json based settings)

    Print the difference.

    If quiet is True, then just return True/False depending if they differ or not.

    The functions assumes that both settings have the same structure.
    """

    if not isinstance(A, Settings):
        print(
            "[red]ERROR[/] compare_module_setting: The first Setting is not a Settings object."
        )
        return None

    if not isinstance(B, Settings):
        print(
            "[red]ERROR[/] compare_module_setting: The second Setting is not a Settings object."
        )
        return None

    output = compare_dict(A, B, "")

    if channels is None:
        channels = list(range(16))

    if not quiet:
        print("Comparing settings:")
        print(f"  a: {A.filename.name}")
        print(f"  b: {B.filename.name}")
        print("")
        if output:
            keylen = max(len(k[0]) for k in output)
            print(f"{'settting':<{keylen}} {'a':>10} {'b':>10}")
            for a, b, c in output:
                skip = True
                for c in channels:
                    if f"[{c}]" in a:
                        skip = False
                        break
                if not skip:
                    print(f"{a:<{keylen}} {b:>10} {c:>10}")
        else:
            print("Settings are identical")

    else:
        return len(output) > 0



import numpy as np
from numpy.testing import assert_array_equal
import pytest

import pixie16


class TestStreamer:
    def test_add_and_read(self):
        S = pixie16.read.StreamReader()

        x = np.arange(20, dtype=np.uint32)

        S.put(x)

        data = S.read(4)
        assert_array_equal(data, np.array([0, 1, 2, 3], dtype=np.uint32))

        data = S.read(4)
        assert_array_equal(data, np.array([4, 5, 6, 7], dtype=np.uint32))

        data = S.read(8)
        assert_array_equal(
            data, np.array([8, 9, 10, 11, 12, 13, 14, 15], dtype=np.uint32)
        )

    def test_read_more_than_available(self):
        S = pixie16.read.StreamReader()

        x = np.arange(20, dtype=np.uint32)

        S.put(x)

        with pytest.raises(pixie16.read.EmptyError):
            S.read(30)

    def test_add_and_read_twice(self):
        S = pixie16.read.StreamReader()

        x = np.arange(20, dtype=np.uint32)

        S.put(x)

        S.read(4)
        assert_array_equal(S.read(4), np.array([4, 5, 6, 7], dtype=np.uint32))

    def test_buffer_size_doubling(self):
        S = pixie16.read.StreamReader(initial_buffer_size=1_000)

        data = np.arange(20, dtype=np.uint32)
        S.put(data)

        assert S.buffer_size == 1_000

        data = np.arange(1_200, dtype=np.uint32)
        S.put(data)

        assert S.buffer_size == 2_000

    def test_buffer_size_adjust_large_data(self):
        S = pixie16.read.StreamReader(initial_buffer_size=100)

        data = np.arange(600, dtype=np.uint32)
        S.put(data)

        assert S.buffer_size == 800

    def test_buffer_size_shrink_size(self):
        S = pixie16.read.StreamReader(initial_buffer_size=100)

        # force increase to 800
        data = np.arange(600, dtype=np.uint32)
        S.put(data)
        assert S.buffer_size == 800

        S.read(600)

        data = np.arange(10, dtype=np.uint32)
        S.put(data)

        assert S.buffer_size == 100
